/* all2all.c  ---by Xiang Yu, Princeton University, June 1998
 *
 * This is a vmmc program that produce following communication pattern:
 * each nodes send to all others beginning from the one next to it.
 *
 * The # of hosts, message size, total messages involved can be changed 
 * in command line. Default message size is 1 word.
 *
 * The nodes involved is consecutive # of nodes in hostfile begining with
 * the node on which the program started.
 */
      

#include "../libSPMD/sbl_support.h"
#include "../libSPMD/spmd_comm.h"
#include "../libSPMD/channel.h"
#include "../libSPMD/timing.h"

#define ID_CHANNEL_BEGIN 100

#define PAGE_ALIGNED_SEND_BUFFER
#define PAGE_ALIGNED_RECEIVE_BUFFER
#define OVERLAP_SEND_BUFFER
#define OVERLAP_RECEIVE_BUFFER

#define CHECK_BUFFER_START

unsigned hosts=2;
unsigned test_size=1;
unsigned channels=2;
unsigned repeat_times=10000;
unsigned warmup_times=200;

/* each proc should have 'channels' sender_buffers and receiver_buffers */
int32 **send_buffers;
int32 **recv_buffers;
PChannel chan;

unsigned All2All(unsigned repeat_times);

void main(int argc, char *argv[])
{
  unsigned i,j,k;
  Cycle start, end;
  unsigned rep;

  /* Check slave and shift slave argument */
  CheckSlaveArg(&argc, argv);

  if (argc<2) {
    fprintf(stderr, "Usage: %s num_hosts [num_words, default=1] [num_msgs, default to 10000]\n", argv[0]);
    exit(0);
  }

  /* set # hosts involved */
  if (atoi(argv[1])>2)
    hosts=atoi(argv[1]);

  if (argc>2)
    if (atoi(argv[2])>0)
      test_size=atoi(argv[2]);

  if (argc>3)
    if (atoi(argv[3])>0)
      repeat_times=atoi(argv[3]);

  /* spawn all other procs on other nodes */
  SpawnSlaves(hosts-1, argc, argv);

  TimingInit();

  fprintf(stderr, "Hello world from a host of %d hosts, message size %d\n",
	  hosts, test_size);


  /* allocate send and receive buffers */
  send_buffers=SafeMalloc(sizeof(int32*)*channels, 
			   "Main: Cannot allocate send buffer index\n");
  recv_buffers=SafeMalloc(sizeof(int32*)*channels*hosts, 
			   "Main: Cannot allocate recv buffer index\n");
  for(i=0; i<channels; i++) {
#ifdef NO_OVERLAP_SEND_BUFFER
    send_buffers[i]=NonOverlapMalloc(sizeof(int32*)*test_size);
#else
#ifdef PAGE_ALIGNED_SEND_BUFFER 
    send_buffers[i]=PageAlignedMalloc(test_size/vmmc_PageSize()+1);
#else
    send_buffers[i]=SafeMalloc(sizeof(int32*)*test_size, 
				"Main: Cannot allocate send buffer\n");
#endif
#endif
//  fprintf(stderr, "Buffer start from %p\n", send_buffers[i]);

    for(j=0; j<hosts; j++) {
#ifdef NO_OVERLAP_RECEIVE_BUFFER
      recv_buffers[j*channels+i]=NonOverlapMalloc(sizeof(int32*)*test_size);
#else
#ifdef PAGE_ALIGNED_RECEIVE_BUFFER
      recv_buffers[j*channels+i]=PageAlignedMalloc(test_size/vmmc_PageSize()+1);
#else
      recv_buffers[j*channels+i]=SafeMalloc(sizeof(int32*)*test_size, 
				  "Main: Cannot allocate recv buffer\n");
#endif
#endif
//    fprintf(stderr, "Buffer start from %p\n", recv_buffers[j*channels+i]);
    }
  }

  /* allocate channel */
  chan=AllocateChannel(hosts, hosts, channels);

  /* Register channels */
  for(i=0;i<hosts;i++)
    for(j=0;j<hosts;j++)
      if (i!=j)
	for(k=0;k<channels;k++) {
	  Channel(chan, i, j, k)=
	    RegisterTransfer(i, send_buffers[k], 
			     j, recv_buffers[i*channels+k], 
			     sizeof(int32)*test_size,
			     ID_CHANNEL_BEGIN+ChannelIndex(chan, i, j, k),
			     -1);
#ifdef CHECK_BUFFER_START
	  recv_buffers[i*channels+k][0]=-1;
#endif
	}
  
  /* finish initialization(import, export), print a message */
  fprintf(stderr, "Hosts %d: channels allocated\n", GetMyProcIndex());
  fflush(stderr);
  
  rep=All2All(warmup_times);

  GETCYCLE(start);
  rep=All2All(repeat_times);
  GETCYCLE(end);

  if (1||IsProc(0)) { 
    fprintf(stderr, "Do work: Average all2all latency=%.2f us, Single node effective bandwidth=%.2f MB/S\n",
	    uInterval(start,end)/rep*(TotalProcs()-1), 
	    test_size*sizeof(int32)*rep/uInterval(start,end));
  }
  else {
    fprintf(stderr, "I am done\n");
  }
}  

unsigned All2All(unsigned repeat_times)
{
  unsigned rep;
  unsigned from, to, c;
  unsigned myid=GetMyProcIndex();

  for(rep=0; rep<repeat_times; rep+=TotalProcs()-1) {
    c=rep%channels; /* switch channel each iteration to avoid race condition*/
    /* send part of the all-all */
    for(to=(myid+1)%TotalProcs(); to!=myid; to=(to+1)%TotalProcs()) {
      send_buffers[c][0]=rep;
      send_buffers[c][test_size-1]=rep;
      DoTransferSend(Channel(chan, myid, to, c), rep, 10);
    }
      
    /* receive part of the all-all */
    for(from=(myid+1)%TotalProcs(); from!=myid; from=(from+1)%TotalProcs()) {
#ifdef CHECK_BUFFER_START
      MonitorWord(Channel(chan, from, myid, c), 0, rep, 10);
#endif
      DoTransferReceive(Channel(chan, from, myid, c), rep, rep, 10);
    }
  }
  return rep;
}
