/* RandomOne2All.c  ---by Xiang Yu, Princeton University June 1998
 *
 * This is a VMMC program that produce following communication pattern:
 * 1. choose master as node a
 * 2. choose node b!=a randomly
 * 3. do a->others and others->b 
 * 4. let a equals to b and go back to 2 until we decide to stop and a is master node again
 *
 * The # of hosts, message size, total messages involved can be changed 
 * in command line. Default message size is 1 word.
 * 
 * The nodes involved is consecutive # of nodes in hostfile begining with
 * the node on which the program started.
 */
      

#include "../libSPMD/sbl_support.h"
#include "../libSPMD/spmd_comm.h"
#include "../libSPMD/channel.h"
#include "../libSPMD/timing.h"

#define ID_CHANNEL_BEGIN 100

#define PAGE_ALIGNED_SEND_BUFFER
#define PAGE_ALIGNED_RECEIVE_BUFFER
#define OVERLAP_SEND_BUFFER
#define OVERLAP_RECEIVE_BUFFER

#define CHECK_BUFFER_START

unsigned hosts=2;
unsigned test_size=1;
unsigned channels=1;
unsigned repeat_times=10000;
unsigned warmup_times=200;
unsigned random_seed=10;

/* each proc should have 'channels' sender_buffers and receiver_buffers */
int32 **send_buffers;
int32 **recv_buffers;
PChannel chan;

unsigned One2All(unsigned repeat_times);

void main(int argc, char *argv[])
{
  unsigned i,j,k;
  Cycle start, end;
  unsigned rep;

  /* Check slave and shift slave argument */
  CheckSlaveArg(&argc, argv);

  if (argc<2) {
    fprintf(stderr, "Usage: %s num_hosts [num_words, default=1] [num_msgs, default to 10000]\n", argv[0]);
    exit(0);
  }


  /* set # hosts involved */
  if (atoi(argv[1])>2)
    hosts=atoi(argv[1]);

  if (argc>2)
    if (atoi(argv[2])>0)
      test_size=atoi(argv[2]);

  if (argc>3)
    if (atoi(argv[3])>0)
      repeat_times=atoi(argv[3]);

  /* spawn all other procs on other nodes */
  SpawnSlaves(hosts-1, argc, argv);

  TimingInit();

  fprintf(stderr, "Hello world from a host of %d hosts, message size %d\n",
	  hosts, test_size);


  /* allocate send and receive buffers */
  send_buffers=SafeMalloc(sizeof(int32*)*channels, 
			   "Main: Cannot allocate send buffer index\n");
  recv_buffers=SafeMalloc(sizeof(int32*)*channels*hosts, 
			   "Main: Cannot allocate recv buffer index\n");
  for(i=0; i<channels; i++) {
#ifdef NO_OVERLAP_SEND_BUFFER
    send_buffers[i]=NonOverlapMalloc(sizeof(int32*)*test_size);
#else
#ifdef PAGE_ALIGNED_SEND_BUFFER 
    send_buffers[i]=PageAlignedMalloc(test_size/vmmc_PageSize()+1);
#else
    send_buffers[i]=SafeMalloc(sizeof(int32*)*test_size, 
				"Main: Cannot allocate send buffer\n");
#endif
#endif
//  fprintf(stderr, "Buffer start from %p\n", send_buffers[i]);

    for(j=0; j<hosts; j++) {
#ifdef NO_OVERLAP_RECEIVE_BUFFER
      recv_buffers[j*channels+i]=NonOverlapMalloc(sizeof(int32*)*test_size);
#else
#ifdef PAGE_ALIGNED_RECEIVE_BUFFER
      recv_buffers[j*channels+i]=PageAlignedMalloc(test_size/vmmc_PageSize()+1);
#else
      recv_buffers[j*channels+i]=SafeMalloc(sizeof(int32*)*test_size, 
				  "Main: Cannot allocate recv buffer\n");
#endif
#endif
//    fprintf(stderr, "Buffer start from %p\n", recv_buffers[j*channels+i]);
    }
  }

  /* allocate channel */
  chan=AllocateChannel(hosts, hosts, channels);

  /* Register transters of each pair of nodes*/
  for(i=0;i<hosts;i++)
    for(j=0;j<hosts;j++)
      if (i!=j) 
	for(k=0;k<channels;k++) {
	  Channel(chan, i, j, k)=
	    RegisterTransfer(i, send_buffers[k], 
			     j, recv_buffers[i*channels+k], 
			     sizeof(int32)*test_size,
			     ID_CHANNEL_BEGIN+ChannelIndex(chan, i, j, k),
			     -1);
#ifdef CHECK_BUFFER_START
	  recv_buffers[i*channels+k][0]=-1;
#endif
	}
  
  /* finish initialization(import, export), print a message */
  fprintf(stderr, "Hosts %d: channels allocated\n", GetMyProcIndex());
  fflush(stderr);
  
  /* initialization the random seed */
  /* IMPORTANT: all procs should have same seed */
  srand(random_seed);

  rep=One2All(warmup_times);
  fprintf(stderr, "I am warm\n");

  GETCYCLE(start);
  rep=One2All(repeat_times);
  GETCYCLE(end);

  if (IsProc(0)) { /* master */
    fprintf(stderr, "Done: Oneway latency=%.2f us,  Bandwidth=%.2f MB/S\n",
	    uInterval(start,end)/rep/channels, 
	    test_size*sizeof(int32)*channels*rep/uInterval(start,end));
  }
  else {
    fprintf(stderr, "I am done\n");
  }
}  

unsigned One2All(unsigned repeat_times)
{
  unsigned rep=0;
  unsigned from, to, next;
  unsigned c;

  from=0;

  do {
    do {
      next=rand()%hosts;
    } while(next==from);

    /* from -> others */
    for(to=0; to<hosts; to++)
      if (to!=from) {
	for(c=0; c<channels; c++) {
	  send_buffers[c][test_size-1]=rep;
#ifdef CHECK_BUFFER_START
	  /* to check buffer start, initialize it first */
	  send_buffers[c][0]=rep;
	  /* receiver check first word(index 0) in reciever buffer, timeout 10 */
	  if (from!=to)
	    MonitorWord(Channel(chan, from, to, c), 0, rep, 10);
#endif
	  DoTransfer(Channel(chan, from, to, c), rep, rep, 10);
	}
	rep++;
      }
	  
    /* others -> next */
    for(to=0; to<hosts; to++)
      if (to!=from && to!=next) {
	/* 
	 * NOTE: we do not do from->next since we just did from->next,
	 *       otherwise reciever can not distinguish two messages
	 */
	for(c=0; c<channels; c++) {
	  send_buffers[c][test_size-1]=rep;
#ifdef CHECK_BUFFER_START
	  /* to check buffer start, initialize it first */
	  send_buffers[c][0]=rep;
	  /* receiver check first word(index 0) in reciever buffer, timeout 10 */
	  if (next!=to)
	    MonitorWord(Channel(chan, to, next, c), 0, rep, 10);
#endif
	  DoTransfer(Channel(chan, to, next, c), rep, rep, 10);
	}
	rep++;
      }

    /* prepare for next iteration */
    from=next;

  } while(rep<repeat_times || from!=0);

  return rep;
}
