/******************************************************************************
 *				all2all.c
 *
 *	Measures cross-section bandwidth with multiple nodes sending to others
 *  at the same time.
 *
 *	syntax:	all2all.exe number_of_nodes number_of_words iterations
 *
 *****************************************************************************/

#include <stdio.h>
#include <windows.h>
#include "vmmc.h"
#include "utils.h"

/******************************************************************************
 *						global macros
 *****************************************************************************/
#define	CSBW_MAX_NODES		16
#define	CSBW_INFO_BUFID		0x00000bad
#define	CSBW_TEST_BUFID(i)	(0x000dad00 + (i))
#define	MYEXIT(code)		exit(code)


typedef struct global_info {
  unsigned int	myNode;
  unsigned int	numNodes;
  vmmc_pid_t	workerPids[CSBW_MAX_NODES];
  vmmc_node_t	nodes[CSBW_MAX_NODES];
  unsigned int	barrier[CSBW_MAX_NODES];
  unsigned int	time[CSBW_MAX_NODES];
  int			errCode;
  unsigned int	syncNumber;
} global_info_t;


/******************************************************************************
 *						global variables
 *****************************************************************************/
unsigned int	num_nodes		= 1;
unsigned int	num_words		= 4;
unsigned int	iterations 		= 10000;

unsigned int	nhosts;
char			methodstring[64]= "random";

vmmc_node_t		*nodelist;

global_info_t	csbwGlobalInfo;

vmmc_pid_t		workerPids[CSBW_MAX_NODES];
vmmc_node_t		nodes[CSBW_MAX_NODES];


unsigned int	numImportedWords[CSBW_MAX_NODES];


int32		*	localDataBuffer;
int32		*	localRecvBuffers[CSBW_MAX_NODES];
int32		*	dataProxies[CSBW_MAX_NODES];
int32		*	infoProxies[CSBW_MAX_NODES];


/******************************************************************************
 *					a simple and inefficient barrier function
 *****************************************************************************/

void barrier(void)
{
  static unsigned int		barrierNumber = 1;
  unsigned int				i, spinCount=0;
  volatile unsigned int *	barPtr;
  global_info_t *			infoPtr;
  int						errc;

  for (i=0; i<csbwGlobalInfo.numNodes; i++) {
    if (i==csbwGlobalInfo.myNode)
      continue;
    DBGPRINT(10,("all2all: sending barrier to node %d ...\n", i));
    infoPtr = (global_info_t*)infoProxies[i];
    errc = vmmc_SendData(&barrierNumber,
			&(infoPtr->barrier[csbwGlobalInfo.myNode]), 1);
    if (errc<0) {
      vmmc_Error(errc, "sending barrier to another worker process");
      exit(1);
    }
  }

  for (i=0; i<csbwGlobalInfo.numNodes; i++) {
    if (i==csbwGlobalInfo.myNode)
      continue;
    barPtr = &csbwGlobalInfo.barrier[i];
    while ( *barPtr < barrierNumber ) {
		if (++spinCount % 0xFFFFFFFF == 0)  {
			DBGPRINT(0, ("Node %d: waiting at the barrier for Node %d\n", 
				csbwGlobalInfo.myNode, i));
		}
    }
  }

  barrierNumber += 0xDEAD;
}




/******************************************************************************
 *				cross-section-bandwidth function
 *****************************************************************************/
void csbw_core(int isMaster)
{
  int						errc;			/* error code */
  unsigned int				i, j, spinCounter = 0;
  volatile unsigned int*	myNodePtr;
  volatile unsigned int*	syncNumberPtr;
  unsigned long 			t1,t2, t3;
  unsigned int				t;
  volatile unsigned int*	timePtr;
  float						bw;
  unsigned int				order[256];
  unsigned int				sendcount=0;
  vmmc_imphandle_t			imphandle;
  vmmc_exphandle_t			exphandle;

  DBGPRINT(0, ("all2all test starting ...\n"));

  if ( isMaster ) {
	goto do_real_work;
  }

  csbwGlobalInfo.myNode   = CSBW_MAX_NODES;
  csbwGlobalInfo.syncNumber = 0xFFFFFFFF;

  myNodePtr   = (volatile unsigned int *) &csbwGlobalInfo.myNode;
  syncNumberPtr = (volatile unsigned int *) &csbwGlobalInfo.syncNumber;

  errc = vmmc_ExportRecvBuf((int32*)&csbwGlobalInfo,
                           sizeof(csbwGlobalInfo)/sizeof(int32),
                           CSBW_INFO_BUFID, NULL, 0, 0, &exphandle);
  if (errc < 0) {
    vmmc_Error(errc, "exporting info buf");
    exit(1);
  }

  spinCounter = 0;
  while ( *syncNumberPtr == 0xFFFFFFFF)
    spinCounter ++;

  DBGPRINT(0, ("all2all: after %d spins, my node id = %d\n",
               spinCounter, *myNodePtr));

  if (csbwGlobalInfo.errCode == -1) {
    DBGPRINT(0, ("all2all: my master told me to quit\n"));
    exit(1);
  }

  /* importing the info buffer from everybody else */
  for (i=0; i<csbwGlobalInfo.numNodes; i++) {
    if (i==csbwGlobalInfo.myNode)
      continue;
    infoProxies[i]	= NULL;
    numImportedWords[i]	= 0;
    errc = vmmc_ImportRecvBuf(
						csbwGlobalInfo.nodes[i],
                        csbwGlobalInfo.workerPids[i],
						CSBW_INFO_BUFID,
						&imphandle,
                        &infoProxies[i],
						&numImportedWords[i]);
    if (errc<0) {
      DBGPRINT(0,("%s: failed to import an info buffer from node %d\n",
                  "all2all", i));
      vmmc_Error(errc, "importing test buffers");
    }
    DBGPRINT(0, ("all2all: imported the info buffer from node %d\n", i));
  }
do_real_work:
  barrier();

  /* importing info buffers from other nodes */
  for (i=0; i<csbwGlobalInfo.numNodes; i++) {
    if (i==csbwGlobalInfo.myNode)
      continue;
    localRecvBuffers[i] = (int32*) malloc(num_words*sizeof(int));
    if (localRecvBuffers[i] == NULL) {
      DBGPRINT(0, ("all2all: failed to allocate %d words for node %d\n",
                   num_words, i));
      continue;
    }
    memset((void*)localRecvBuffers[i], 0, sizeof(int)*num_words);
    errc = vmmc_ExportRecvBuf(localRecvBuffers[i], num_words,
                             CSBW_TEST_BUFID(i), NULL, 0, 0, &exphandle);
    if (errc<0) {
      DBGPRINT(0,("all2all: failed to export a test buffer for node %d\n",
                  i));
      vmmc_Error(errc, "exporting test buffers");
    }
  }

  localDataBuffer = (int32*) malloc(num_words*sizeof(int));
  if (localDataBuffer == NULL) {
    DBGPRINT(0, ("all2all: failed to allocate %d words for local data buffer\n",
                 num_words));
    exit(1);
  }
  memset((void*)localDataBuffer, 0, sizeof(int32)*num_words);

  barrier();

  for (i=0; i<csbwGlobalInfo.numNodes; i++) {
    if (i==csbwGlobalInfo.myNode)
      continue;
    dataProxies[i]	= 0;
    numImportedWords[i]	= 0;
    errc = vmmc_ImportRecvBuf(
				csbwGlobalInfo.nodes[i],
                csbwGlobalInfo.workerPids[i],
				CSBW_TEST_BUFID(csbwGlobalInfo.myNode),
				&imphandle,
				&dataProxies[i],
				&numImportedWords[i]
			);
    if (errc<0) {
      DBGPRINT(0,("%s: failed to import a test buffer from node %d\n",
                  "all2all", i));
      vmmc_Error(errc, "importing test buffers");
    }
  }

  /* global sync */
  barrier();

  //DBGPRINT(0, ("%s: global synchronizaion is done\n", "all2all"));

  // establish a random order for sends
  srand(GetTickCount()+ 10117*csbwGlobalInfo.myNode);
  for (i=0; i<csbwGlobalInfo.numNodes; i++) {
	  if (strcmp(methodstring, "pairwise") == 0) {
		order[i] = csbwGlobalInfo.numNodes-1-csbwGlobalInfo.myNode	;
	  } else {
		order[i] = (i+csbwGlobalInfo.myNode) % csbwGlobalInfo.numNodes;
	  }
  }
	if (strcmp(methodstring, "random") == 0) {
	  for (i=0; i<1000; i++) {
		  int	i1, i2;
		  int	t;

		  i1 = rand() % csbwGlobalInfo.numNodes;
		  i2 = rand() % csbwGlobalInfo.numNodes;

		  // exchange these two indices
		  t = order[i1];
		  order[i1] = order[i2];
		  order[i2] = t;
	  }
	}
	  


  barrier();
  barrier();


  t1 = GetTickCount();
  sendcount = 0;
  for (j=1; j<=iterations; j++) {
    /* DBGPRINT(0,("%s: iterations %d ...\n", "all2all", j)); */

    for (i=0; i<csbwGlobalInfo.numNodes; i++) {
		unsigned int	k;

		k = order[i];
		if ( k == csbwGlobalInfo.myNode)
			continue;
		sendcount ++;
        if (numImportedWords[k] != 0) {
         DBGPRINT(1,("%s: sending data to node %d ...\n", "all2all", k));
         errc = vmmc_SendData(localDataBuffer, dataProxies[k], num_words);
         if (errc<0) {
           vmmc_Error(errc, "sending data to another worker process");
         }
		}
    }
  }
  t2 = GetTickCount();

  barrier();
 
  t3 = GetTickCount();
  t   = (t2-t1)*1000;
  bw  = (float)(num_words*sendcount*sizeof(int));
  memset(&csbwGlobalInfo.time[0], 0, sizeof(csbwGlobalInfo.time));
 
  barrier();


  if (! isMaster ) {
    DBGPRINT(0, ("CWBW Bandwidth = %6f MB/sec\n", bw/(float)t));
    errc = vmmc_SendData((int32*)&t,
		&(((global_info_t*)infoProxies[0])->
				time[csbwGlobalInfo.myNode]),
			sizeof(t));
    if (errc<0) {
      vmmc_Error(errc, "sending bandwidth to master node");
    }
  } else {
    DBGPRINT(0,
	("CWBW Bandwidth for Node %d = %6f MB/sec\n", 0, bw/(float)t));
    for(i=1; i<csbwGlobalInfo.numNodes; i++) {
      timePtr = (volatile unsigned int*) & csbwGlobalInfo.time[i];
      while (*timePtr == 0) {
        spinCounter++;
      }
	  Sleep(1000);
	  t = *timePtr;
      DBGPRINT(0, 
	("CSBW Bandwidth for Node %d = %6f\n", i, bw/(float)t));
    }
  }

  DBGPRINT(0, ("CSBW [%s] %d words %d iterations Overall Bandwidth = %f\n",
			methodstring, num_words, iterations,
			bw*csbwGlobalInfo.numNodes/((t3-t1)*1000)));

  barrier();

  Sleep(2000);	/* make sure other nodes get the messages */
  exit(0);
}





/******************************************************************************
 *			main: spawn the workers
 *****************************************************************************/
#define	CSBW_NUM_ARGS	4

int main(int argc, char *argv[])
{
  vmmc_result_t		errc;
  uint32			i, actualNodeCount;
  int				numErrors = 0;
  char *			slargv[5];
  vmmc_exphandle_t	exphandle;
  vmmc_imphandle_t	imphandle;

  SET_DEBUG_LEVEL(0);

  if (argc < (CSBW_NUM_ARGS+1) ) {
    DBGPRINT(0, ("%s [random|pairwise] num_nodes num_words iterations\n", argv[0]));
    exit(0);
  }
  strcpy(methodstring, argv[1]);
  num_nodes  = atoi(argv[2]);
  num_words  = atoi(argv[3]);
  iterations = atoi(argv[4]);

  memset((void*)&csbwGlobalInfo, 0, sizeof(csbwGlobalInfo));


  if ((argc==(CSBW_NUM_ARGS+2)) && strcmp(argv[CSBW_NUM_ARGS+1], "slave")==0) {
    csbw_core(0);
    exit(0);
  }

  /* master */
  if (argc != (CSBW_NUM_ARGS+1)) {	/* master fromat */
    DBGPRINT(0, ("%s bad format: %s num_nodes num_words iterations\n",
		 "all2all", argv[0]));
    exit(1);
  }

  /* print machine configuration */
  errc = vmmc_SessionHosts(&nhosts, &nodelist);
  if (errc < 0) {
    vmmc_Error(errc, "getting hosts");
    exit(1);
  }
  DBGPRINT(0, ("%s Num hosts = %d\n", "all2all", nhosts));
  for (i = 0; i < nhosts; i++) {
    DBGPRINT(0, ("%s Host number %d, host address %s\n",
		 "all2all", i, vmmc_NodeToName(*(nodelist + i))));
  }

  if (nhosts < num_nodes) {
    DBGPRINT(0, ("%s: not enough nodes in this configuration\n", argv[0])); 
    exit(1);
  }

  /* prepare slave args */
  slargv[0] = argv[1];
  slargv[1] = argv[2];
  slargv[2] = argv[3];
  slargv[3] = argv[4];
  slargv[4] = "slave";
  slargv[5] = NULL;
 
  DBGPRINT(0, ("%s: spawning %d test processes\n", "all2all", num_nodes));

  actualNodeCount = 0;
  for (i=0; i<nhosts; i++) {
    if (vmmc_EqualNodes(vmmc_MyNode(), nodelist[i])) {
      csbwGlobalInfo.nodes[actualNodeCount]		= nodelist[i];
      csbwGlobalInfo.workerPids[actualNodeCount]	= vmmc_MyPid();
      actualNodeCount ++;
      break;
    }
  }

  for (i=0; i<nhosts; i++) {
    if (actualNodeCount == num_nodes) break;

    if (vmmc_EqualNodes(vmmc_MyNode(), nodelist[i])) {
      continue;
    }
    csbwGlobalInfo.nodes[actualNodeCount] = nodelist[i];
    errc = vmmc_Spawn(argv[0], slargv,
		     nodelist[i], &csbwGlobalInfo.workerPids[actualNodeCount]);
    if (errc < 0) {
      vmmc_Error(errc, "spawning slaves");
      exit(0);
    } else {
      DBGPRINT(0, ("%s: slave no. %2d is spawned, pid = %4u on %s\n",
         "all2all", actualNodeCount, csbwGlobalInfo.workerPids[actualNodeCount],
         vmmc_NodeToName(nodelist[i])));
      actualNodeCount ++;
    }
  }

  if (actualNodeCount < num_nodes) {
    DBGPRINT(0,("%s: only spawned %d test processes, let us die!\n",
             "all2all", actualNodeCount));
    exit(1);
  }

  csbwGlobalInfo.numNodes = actualNodeCount;

  errc = vmmc_ExportRecvBuf((int32*)&csbwGlobalInfo,
                           (uint32)sizeof(csbwGlobalInfo)/sizeof(uint32),
                           CSBW_INFO_BUFID,
						   NULL,
						   0,
						   0,
						   &exphandle);
  if (errc < 0) {
    vmmc_Error(errc, "exporting info buf");
    exit(1);
  }


  numErrors = 0;
  for (i=1; i<actualNodeCount; i++) {
    infoProxies[i]	= 0;
    numImportedWords[i]	= 0;
    DBGPRINT(0,
	("%s: master: importing the info buffer from node %d\n", argv[0], i));
    errc = vmmc_ImportRecvBuf(
							csbwGlobalInfo.nodes[i],
							csbwGlobalInfo.workerPids[i],
							CSBW_INFO_BUFID,
							&imphandle,
							&infoProxies[i],
							&numImportedWords[i]
						);
    if (errc<0) {
      DBGPRINT(0,("%s: failed to import a buffer from slave no. %d\n",
                  "all2all", i));
      vmmc_Error(errc, "importing children buffers");
      numErrors ++;
    }
  }

  if (numErrors)
    csbwGlobalInfo.errCode = -1;
  else
    csbwGlobalInfo.errCode = 0;

  csbwGlobalInfo.syncNumber = 1;
  for (i=1; i<actualNodeCount; i++) {
    if (numImportedWords[i] == 0)
      continue;
    csbwGlobalInfo.myNode = i;
    errc = vmmc_SendData((int32*)&csbwGlobalInfo, infoProxies[i],
			sizeof(csbwGlobalInfo)/sizeof(int32));
    if (errc<0) {
      vmmc_Error(errc, "sending global info to worker process");
    }
  }

  csbwGlobalInfo.myNode = 0;

  DBGPRINT(0,("%s: successfully sent the global information to all workers\n",
              "all2all") );

  if (csbwGlobalInfo.errCode==-1) {
    DBGPRINT(0, ("%s: quiting due to errors\n", "all2all"));
    exit(0);
  }

  csbw_core(1);
  exit(0);
  return 0;
}
