/******************************************************************************
 *						bandwidth.c
 *
 *	Measures two-node bandwidth for multiple message sizes.
 *
 *****************************************************************************/


#include <stdio.h>
#include <windows.h>
#include "vmmc.h"
#include "utils.h"


uint32	iterations;
uint32	num_words;
uint32	num_pages;
uint32	nbytes;
uint32	nwords, start_words;

char *SlaveName  = "bandwidth: Slave ";
char *MasterName = "bandwidth: Master";




#define	BANDWIDTH_BUFFER_ID		222



/*
 *	int32 *	allocbuf(nwords):
 *
 *		allocate a word-aligned buffer for nwords
 */
int32* allocbuf(unsigned int nwords)
{
	unsigned long	bufaddr;

	bufaddr = (unsigned long)malloc(sizeof(int32)*(nwords+1));
	if (bufaddr == 0) {
		fprintf(stderr, "allocbuf: failed to allocate %d bytes\n",
				sizeof(int32)*(nwords+1));
		exit(0);
	}

	memset((void*)bufaddr, 0, sizeof(int32)*(nwords+1));
	if ( bufaddr & (sizeof(int32)-1) ) {
		bufaddr += sizeof(int32);
		bufaddr &= ~(sizeof(int32)-1);
	}
	return (int32*) bufaddr;
}




/*
 *	runtest:
 *
 *		Both master and slave calls this function.
 */
void
runtest(int master, vmmc_node_t remoteNode, vmmc_pid_t remotePid)
{
	vmmc_result_t		errc;			/* error code */
	vmmc_exphandle_t	exphandle;
	vmmc_imphandle_t	imphandle;
	uint32				i;
	int32 *				dstBuf;
	int32 *				srcBuf;
	volatile int32 *	srcbufend;
	volatile int32 *	dstbufend;
	int32 *				proxy;
	unsigned long		spinctr;

	printf("%s test started ...\n", "bandwidth:");

	srcBuf = allocbuf(num_words);
	dstBuf = allocbuf(num_words);
	srcbufend = &(srcBuf[num_words - 1]);
	dstbufend = &(dstBuf[num_words - 1]);

	for(i=0; i<num_words; i++) {
		dstBuf[i]=0;
		srcBuf[i]=0;
	}

	DBGPRINT(0, ("bandwidth: srcbuf=0x%x dstbuf=0x%x\n",
			(int)srcBuf, (int)dstBuf));


	errc = vmmc_ExportRecvBuf(dstBuf, num_words, BANDWIDTH_BUFFER_ID,
							  NULL, BUF_WRITE_ONLY, 0, &exphandle);
	if (errc < 0) {
		vmmc_Error(errc, "exporting receive buffer");
		exit(1);
	}

	errc = vmmc_ImportRecvBuf(remoteNode, remotePid, BANDWIDTH_BUFFER_ID,
							&imphandle,	&proxy, &nwords);
	if (errc < 0) {
		vmmc_Error(errc, "importing remote buffer");
		exit(1);
	}
	DBGPRINT(2, ("bandwidth: imported %d words at proxy=0x%x\n", 
				nwords, (unsigned int)proxy));

	/*
	 *	The virtual-to-physical translation table is set up on demand for the
	 *	first send. Let's send once to fact out the time for translation setup.
	 */
	if ((errc = vmmc_SendData(srcBuf, proxy, num_words)) < 0) {
		vmmc_Error(errc, "send data to each other");
		exit(1);
	}

	for(; start_words<=num_words; start_words+=start_words) {
		DWORD	t1, t2;

		srcbufend = &(srcBuf[start_words - 1]);
		dstbufend = &(dstBuf[start_words - 1]);
		nbytes = start_words * sizeof(int32);
    
		t1 = GetTickCount();	/* returns time in milliseconds */
		for (i = 1; i <= iterations; i++) {
			if ( ! master ) {
				while (*dstbufend != (volatile int32) i)
					spinctr++;
			}
			*srcbufend = i;
			if ((errc = vmmc_SendData(srcBuf, proxy, start_words)) < 0) {
				vmmc_Error(errc, "send msg to master");
				exit(1);
			}
			if ( master ) {
				while (*dstbufend != (volatile int32) i)
					spinctr++;
			}
		}


		t2 = GetTickCount();

		DBGPRINT(0, ("\tniter=%d nbytes=%d OneWayLat=%7.3f (us) Band=%7.3f (MB/s)\n",
				iterations,
				nbytes,
				(t2-t1)*1000.0/2.0/iterations,
				2.0*nbytes*iterations/((t2-t1)*1000.0)
		));
	}
	DBGPRINT(0, ("bandwidth: exiting...spinctr=%d\n", spinctr));
}





void
main(int argc, char *argv[])
{
	vmmc_pid_t		remotePid;
	vmmc_node_t		remoteNode;
	vmmc_result_t	errc;
	uint32			i;
	char *			slaveArgv[10];
	uint32			nhosts;
	vmmc_node_t *	nodeList;

	if (argc < 3) {
		fprintf(stderr, "%s num_words iterations [start_nwords]\n", argv[0]);
		exit(0);
	}


	num_words  = atoi(argv[1]);
	iterations = atoi(argv[2]);
	if (argc>=4)
		start_words = atoi(argv[3]);
	else 
		start_words = num_words;

	nbytes    = num_words * sizeof(int32);

	errc = vmmc_Parent(&remoteNode, &remotePid);

	if (errc != vmmcErr_NoParent) { /* the slave */
		runtest(0, remoteNode, remotePid);
		exit(0);
	}


	/* print machine configuration */
	errc = vmmc_SessionHosts(&nhosts, &nodeList);
	if (errc < 0) {
		vmmc_Error(errc, "getting hosts");
		exit(1);
	}

	printf("%s number of hosts = %d\n", "bandwidth:", nhosts);
	for (i = 0; i < nhosts; i++) {
		printf("%s host # %2d: %s\n",
				"bandwidth:", i, vmmc_NodeToName(*(nodeList + i)));
	}

	/* prepare slave args */
	slaveArgv[0] = argv[1];
	slaveArgv[1] = argv[2];
	slaveArgv[2] = argv[3];	/* this one could be null */
	slaveArgv[3] = NULL;

	DBGPRINT(0, ("bandwidth: spawning a slave process ......\n"));
	if (nhosts > 1) {
		if (vmmc_EqualNodes(nodeList[0], vmmc_MyNode())) {
			remoteNode = nodeList[1];
		} else {
			remoteNode = nodeList[0];
		}
	} else {
		remoteNode = nodeList[0];
	}

	errc = vmmc_Spawn(argv[0], slaveArgv, remoteNode, &remotePid);
	if (errc < 0) {
		vmmc_Error(errc, "spawning slaves");
		exit(1);
	}

	DBGPRINT(0, ("bandwidth: spawned slave process (pid=%u) on %s\n",
				remotePid, vmmc_NodeToName(remoteNode)));

	runtest(1, remoteNode, remotePid);
}
