/*****************************************************************************
 *
 *							latency.c
 *
 *		This program measures the one-way communication latency of VMMC by
 *	pingpong a short message between two VMMC nodes.
 *
 *	syntax: latency.exe number_of_words iterations
 *
 *****************************************************************************/

#include <windows.h>
#include <stdio.h>
#include "vmmc.h"
#include "utils.h"


/*****************************************************************************
 *					some useful macros
 *****************************************************************************/
#define	LATENCY_BUFFER_ID	4




/*
 *	int32 *	allocbuf(nwords):
 *
 *		allocate a word-aligned buffer for nwords
 */
int32* allocbuf(unsigned int nwords)
{
	unsigned long	bufaddr;

	bufaddr = (unsigned long)malloc(sizeof(int32)*(nwords+1));
	if (bufaddr == 0) {
		fprintf(stderr, "allocbuf: failed to allocate %d bytes\n",
				sizeof(int32)*(nwords+1));
		exit(0);
	}

	memset((void*)bufaddr, 0, sizeof(int32)*(nwords+1));
	if ( bufaddr & (sizeof(int32)-1) ) {
		bufaddr += sizeof(int32);
		bufaddr &= ~(sizeof(int32)-1);
	}
	return (int32*) bufaddr;
}


/*
 *	runtest:
 */
void runtest(unsigned int nwords, unsigned int iterations,
			 vmmc_node_t remoteNode, vmmc_pid_t remotePid, int slave)
{
	int32*				dstBuf;
	int32*				srcBuf;
	volatile int32		*srcbufend;
	volatile int32		*dstbufend;
	vmmc_result_t		errc;			/* error code */
	unsigned int		i = 0;
	int32 *				proxy;
	unsigned long		spinCount;
	LARGE_INTEGER		t1, t2, freq;
	unsigned long		t;
	unsigned int		nwimp;
	vmmc_imphandle_t	imphandle;
	vmmc_exphandle_t	exphandle;

	DBGPRINT(10, ("Starting latency runtest():\n"));
	DBGPRINT(10, ("\tnwords=%d   iterations=%d\n", nwords, iterations));
	DBGPRINT(10, ("\tremotePid=%u remoteNode=(%s)\n",
				remotePid, vmmc_NodeToName(remoteNode)));

	/* first, allocate the buffer */
	srcBuf = allocbuf(nwords);
	dstBuf = allocbuf(nwords);
	srcbufend = &(srcBuf[nwords - 1]);
	dstbufend = &(dstBuf[nwords - 1]);
	
	DBGPRINT(10, ("\tSource buffer at 0x%x\n", (int)&srcBuf[0]));
	DBGPRINT(10, ("\tFlag word address = 0x%x\n",   (int)srcbufend));
	DBGPRINT(10, ("\tDestination buffer at 0x%x\n", (int)&dstBuf[0]));
	DBGPRINT(10, ("\tFlag word address = 0x%x\n",   (int)dstbufend));


	*dstbufend = 0xFFFFFFFF;

	/*
	 *	Export and Import buffers:
	 */
	errc = vmmc_ExportRecvBuf(
						dstBuf,		/* local recv buffer */
						nwords,
						LATENCY_BUFFER_ID,
						NULL,		/* call-back function */
						0,			/* flags */
						0,			/* perm  */
						&exphandle	/* out */
				);
	if (errc < 0) {
		fprintf(stderr, "latency: ExportRecvBuf() failed: %s\n",
				vmmc_ErrorStr(errc));
		exit(1);
	}
	DBGPRINT(2, ("\tExported destination buffer with %d words\n",
			nwords+1));

	errc = vmmc_ImportRecvBuf(
					remoteNode,
					remotePid,
					LATENCY_BUFFER_ID,
					&imphandle,			/* out */
					&proxy,				/* out */
					&nwimp				/* out */
			);
	if (errc < 0) {
		fprintf(stderr, "latency: ImportRecvBuf() failed: %s\n",
				vmmc_ErrorStr(errc));
		exit(1);
	}
	DBGPRINT(2, ("\tImported remote buffer: nwords=%d, proxy=0x%x\n",
				nwords, (int)proxy));


	/*
	 *	now, the pingpong loop
	 */
	for (i = 0; i <= iterations; i++) {
		if (slave) {
			while (*dstbufend != i) {
				if (spinCount < 0xFFFFFFFF)
					spinCount ++;
			};
		}

		*srcbufend = i;
		if ((errc = vmmc_SendData(srcBuf, proxy, nwords)) < 0) {
			fprintf(stderr, "latency: vmmc_SendData() failed: %s\n",
					vmmc_ErrorStr(errc));
			exit(1);
		}

		if ( ! slave) {
			while (*dstbufend != i) {
				if (spinCount < 0xFFFFFFFF)
					spinCount ++;
			}
		}

		/* first iteration merely sets up the path, we ignore it */
		if (i==0) {
			QueryPerformanceCounter(&t1);
		}

		DBGPRINT(10, ("\t\titeration #%d\n", i));
	}

	QueryPerformanceCounter(&t2);

	QueryPerformanceFrequency(&freq);

	t = (unsigned long)(t2.QuadPart - t1.QuadPart);
	t /= (unsigned long)(freq.QuadPart/1000);

	DBGPRINT(0, ("OneWayLat=%7.3f usec\n", t*1000.0/2.0/iterations));
	DBGPRINT(0, ("total spin count = %u\n", spinCount));
	DBGPRINT(0, ("latency test exiting...\n"));
	exit(0);
}






/*
 *	main:
 *
 *		Parse the command line, spawn the slave process.
 */
void main(int argc, char **argv)
{
	
	vmmc_pid_t		remotePid;
	vmmc_node_t *	hids;
	vmmc_node_t		remoteNode, mynode;
	vmmc_result_t	errc;
	int				i;
	char *			slaveArgv[4];
	int				nhosts;
	unsigned int	iterations;
	unsigned int	nwords;

	SET_DEBUG_LEVEL(0);
	if (argc != 3) {
		fprintf(stderr, "usage: %s nwords iterations\n", argv[0]);
		exit(1);
	}

	nwords     = atoi(argv[1]);
	iterations = atoi(argv[2]);

	errc = vmmc_Parent(&remoteNode, &remotePid);
	/* errc = either vmmcErr_NoParent or vmmc_Success */
	if (errc != vmmcErr_NoParent) { /* the slave */
		runtest(nwords, iterations, remoteNode, remotePid, 1);
		exit(0);
	}

	//printf("vmmc_Parent: errc=%i, remoteNode=%x, remotePid=%i\n", errc, *(int*)&remoteNode, remotePid);
	
	/* print machine configuration */
	errc = vmmc_SessionHosts(&nhosts, &hids);
	if (errc < 0) {
		fprintf(stderr, "latency: vmmc_SessionHosts() failed: %s\n",
				vmmc_ErrorStr(errc));
		exit(1);
	}
	for (i = 0; i < nhosts; i++) {
		DBGPRINT(0, ("\thost[%2d] : %s\n", i, vmmc_NodeToName(hids[i])));
	}


	/*
	 *	select a remote host
	 */
	mynode = vmmc_MyNode();
	if (nhosts > 1) {
		if (vmmc_EqualNodes(hids[0], mynode)) {
		remoteNode = hids[1];
		} else {
			remoteNode = hids[0];
		}
	} else {
		remoteNode = hids[0];
	}



	/* prepare slave args */
	slaveArgv[0] = argv[1];
	slaveArgv[1] = argv[2];
	slaveArgv[2] = NULL;

	DBGPRINT(0, ("%s: spawning slave\n", argv[0]));

	errc = vmmc_Spawn(argv[0], slaveArgv, remoteNode, &remotePid);
	if (errc < 0) {
		fprintf(stderr, "latency: vmmc_Spawn() failed: %s\n",
				vmmc_ErrorStr(errc));
		exit(1);
	}

	DBGPRINT(0, ("%s: slave spawned, slaveSquid= %u on %s\n",
	       argv[0], remotePid, vmmc_NodeToName(remoteNode)));



	runtest(nwords, iterations, remoteNode, remotePid, 0);
}
