#include "spmd_comm.h"

typedef struct _TTransfer {
  unsigned sender_index, receiver_index;
  unsigned buf_id;
  void *sender_buffer;
  volatile int32 *sender_proxy_buffer;
  volatile int32 *receiver_buffer;
  unsigned nwords;
  int32 last_init_value;
  unsigned sender_async:1;
  vmmc_exphandle_t exphandle;
  vmmc_imphandle_t imphandle;
  
  /* No notification handler and redirection yet */
} TTransfer;

static TTransfer invalid_transfer={ 0xffff, 0xffff, 0, NULL, NULL, NULL, 0, 0, 0};
PTransfer INVALID_TRANSFER=&invalid_transfer;

PTransfer RegisterTransfer(unsigned sender_index, void *sender_buffer,
			   unsigned receiver_index, void *receiver_buffer,
			   unsigned buffer_size, unsigned transfer_id, 
			   int32 init_value)
{
  /* IMPORTANT: init_value should have the same value 
   *  on both sender and receiver
   * Maybe also for all other auguments
   */
  if (sender_index==receiver_index && IsProc(sender_index)) {
    int errc;
    PTransfer t; 
    unsigned nwords;
    /* check argument */
    if (buffer_size%vmmc_WordSize()!=0) {
      fprintf(stderr, "RegisterTransfer: buffer_size is not multiple of words\n");
      exit(1);
    }

    t=(PTransfer)malloc(sizeof(TTransfer));
    if (t==NULL) {
      fprintf(stderr, "RegisterTransfer: can not allocate TTransfer structure \n");
      exit(1);
    }
    
    t->sender_index=sender_index;
    t->receiver_index=receiver_index;
    t->buf_id=transfer_id;
    t->sender_buffer=sender_buffer;
    t->receiver_buffer=(volatile int32*)receiver_buffer;
    t->nwords=buffer_size/vmmc_WordSize();
    t->last_init_value=init_value;
    t->sender_proxy_buffer=NULL;
    t->sender_async=0;

    /* init buffer_end value */
    t->receiver_buffer[t->nwords-1]=init_value;
    
    /* receiver export buffer */
    errc=vmmc_ExportRecvBuf(receiver_buffer, t->nwords, transfer_id, NULL, 0, 0, &(t->exphandle));
    if (errc<0) {
      vmmc_Error(errc, "RegisterTransfer: Receiver exporting recv buffer");
      exit(1);
    }

    /* sender import recv buffer */ 
    errc=vmmc_ImportRecvBuf(GetNode(receiver_index), GetSquid(receiver_index), 
			   transfer_id, &(t->imphandle), &(t->sender_proxy_buffer), 
			   &nwords);
    if (errc<0) {
      vmmc_Error(errc, "RegisterTransfer: Sender importing recv buffer");
      exit(1);
    }
    if (nwords!=t->nwords) {
      fprintf(stderr, "RegisterTransfer: sender import nwords is not same as registered\n");
      exit(1);
    }
    return t;
  }

  if (IsProc(sender_index)) { /* this is the sender */
    int errc;
    PTransfer t; 
    unsigned nwords;

    /* check argument */
    assert(sender_index!=receiver_index);

    if (buffer_size%vmmc_WordSize()!=0) {
      fprintf(stderr, "RegisterTransfer: buffer_size is not multiple of words\n");
      exit(1);
    }

    /* allocate buffers */
    t=(PTransfer)malloc(sizeof(TTransfer));
    if (t==NULL) {
      fprintf(stderr, "RegisterTransfer: can not allocate TTransfer structure \n");
      exit(1);
    }
    
    t->sender_index=sender_index;
    t->receiver_index=receiver_index;
    t->buf_id=transfer_id;
    t->sender_buffer=sender_buffer;
    t->receiver_buffer=(int32*)receiver_buffer;
    t->nwords=buffer_size/vmmc_WordSize();
    t->last_init_value=init_value;
    t->sender_async=0;

    /* sender import recv buffer */ 
    errc=vmmc_ImportRecvBuf(GetNode(receiver_index), GetSquid(receiver_index), 
			   transfer_id, &(t->imphandle), &(t->sender_proxy_buffer), 
			   &nwords);
    if (errc<0) {
      vmmc_Error(errc, "RegisterTransfer: Sender importing recv buffer");
      exit(1);
    }
    if (nwords!=t->nwords) {
      fprintf(stderr, "RegisterTransfer: sender import nwords is not same as registered\n");
      exit(1);
    }
    return t;
  }	    
  else if (IsProc(receiver_index)) { /* this is the receiver */
    int errc;
    PTransfer t;

    assert(sender_index!=receiver_index);

    /* check argument */
    if (buffer_size%vmmc_WordSize()!=0) {
      fprintf(stderr, "RegisterTransfer: buffer_size is not multiple of words\n");
      exit(1);
    }

    t=(PTransfer)malloc(sizeof(TTransfer));
    if (t==NULL) {
      fprintf(stderr, "RegisterTransfer: can not allocate TTransfer structure \n");
      exit(1);
    }
    
    t->sender_index=sender_index;
    t->receiver_index=receiver_index;
    t->buf_id=transfer_id;
    t->sender_buffer=sender_buffer;
    t->receiver_buffer=(volatile int32*)receiver_buffer;
    t->nwords=buffer_size/vmmc_WordSize();
    t->last_init_value=init_value;
    t->sender_proxy_buffer=NULL;
    t->sender_async=0;

    /* init buffer_end value */
    t->receiver_buffer[t->nwords-1]=init_value;
    
    /* receiver export buffer */
    /* FIXME: should use vmmc_Export call instead */
    errc=vmmc_ExportRecvBuf(receiver_buffer, t->nwords, transfer_id, NULL, 0, 0, &(t->exphandle));
    if (errc<0) {
      vmmc_Error(errc, "RegisterTransfer: Receiver exporting recv buffer");
      exit(1);
    }
    return t;
  }
  else /* for other nodes: return NULL */
    return NULL; 
}

void DoTransfer(PTransfer t, int32 expected_value, int32 new_init_value, unsigned recv_timeout)
{
  /* IMPORTANT: expected_value and new_init_value should have the same value 
   *  on both sender and receiver
   */

  if (t==NULL)
    return;
  if (IsProc(t->sender_index)) { /* this is the sender */
    Cycle start, end;
    int errc;
    vmmc_async_handle_t msgid;
    int loops;

    if (t->sender_async) {
      fprintf(stderr, "DoTransfer: sender: last async has not been finished\n");
      exit(1);
    }
      
    if (expected_value!=((int32*)t->sender_buffer)[t->nwords-1]) {
      fprintf(stderr, "DoTransfer: Sender: expected_value and buffer-end inconsistent\n");
      exit(1);
    }
    if (t->last_init_value==expected_value) {
      fprintf(stderr, "DoTransfer: Sender: init_value is the same as expected_value\n");
      exit(1);
    }
#if 0   /* sync send */
    errc=vmmc_SendData(t->sender_buffer, t->sender_proxy_buffer, t->nwords);  
#else   /* async send */
    errc=vmmc_SendDataAsync(t->sender_buffer, t->sender_proxy_buffer, t->nwords, &msgid);

    if (errc<0) {
      fprintf(stderr, "DoTransfer: Sending error %x: \n", errc);
      vmmc_Error(errc, "DoTransfer: Sending error");
      fflush(stderr);
      exit(1);
    }

    /* check for send timeout when using async send */
    GETCYCLE(start);
    loops=0;
    do {
      GETCYCLE(end);
      if (recv_timeout!=0 && Interval(start,end)>(float)recv_timeout) {
	fprintf(stderr, "DoTransfer: Sender: time out after %d seconds, sending %i to host %d\n",
		recv_timeout, expected_value, t->receiver_index);
		fflush(stderr);
	GETCYCLE(start);
      }      
      errc=vmmc_AsyncStatus(msgid);
      loops++;
    } while(errc==vmmcErr_NoSuchError);
#if 0 
    /* this is to check if sending is extremely slow: less than 200 words uses more than 10ms */
    if (mInterval(start,end)>10 && t->nwords<200) {
      fprintf(stderr, "DoTransfer: Sending packet %d(%d words) to host %d takes at least %.2f ms\n", 
	      expected_value, t->nwords, t->receiver_index, mInterval(start,end));
    }
#endif
#endif /* async send */
    if (errc<0) {
      fprintf(stderr, "DoTransfer: Sending error1 %x on sending %i, msgid %i, loops %i\n", errc, expected_value, *(int*)&msgid, loops);
      vmmc_Error(errc, "DoTransfer: Sending error1");
      fflush(stderr);
      while((errc=vmmc_AsyncStatus(msgid))<0)
        fprintf(stderr, "DoTransfer: Sending error1 %x on sending %i, msgid %i, loops %i\n", errc, expected_value, *(int*)&msgid, loops);
#if 0      
      exit(1);
#endif
    }
  }

  if (IsProc(t->receiver_index)) { /* this is the receiver */
    int32 v;
    Cycle start, end;
    
    GETCYCLE(start);

#ifndef INIT_VALUE_CHECK
    /* assume the init value is the first one in the buffer we get, but not expected value */
    v=t->receiver_buffer[t->nwords-1];
    if (v!=expected_value)
      t->last_init_value=v;
#else
    if (t->last_init_value==expected_value) {
      fprintf(stderr, "DoTransfer: Receiver: init_value is the same as expected_value\n");
	  fflush(stderr);
      exit(1);
    }
#endif

    do {
      v=t->receiver_buffer[t->nwords-1];
      GETCYCLE(end);
      if (recv_timeout!=0 && Interval(start,end)>(float)recv_timeout) {
	fprintf(stderr, "DoTransfer: Receiver: time out after %d seconds, waiting %i from host %d\n",
		recv_timeout, expected_value, t->sender_index);
	  fflush(stderr);
	GETCYCLE(start);
      }

      if (v!=expected_value && v!=t->last_init_value) { /* get a third value */
	fprintf(stderr, "DoTransfer: Receiver: get a value(0x%x) other than expected(0x%x) or init(0x%x) from host %d\n", v, expected_value, t->last_init_value, t->sender_index);
	  fflush(stderr);
	t->last_init_value=v;
      } 
    } while(v!=expected_value && v==t->last_init_value);
    
    if (v!=expected_value)
      if (v!=t->last_init_value) { /* get a third value */
	fprintf(stderr, "DoTransfer: Receiver: get a value(0x%x) other than expected(0x%x) or init(0x%x)\n", v, expected_value, t->last_init_value);
	  fflush(stderr);
	exit(1);
      } 
      else { /* receiver time out */
	assert(0);
	fprintf(stderr, "DoTransfer: Receiver: time out after %d seconds\n", recv_timeout);
	  fflush(stderr);
	exit(1);
      }	
    
    /* get expected value */
    t->receiver_buffer[t->nwords-1]=new_init_value;
    t->last_init_value=new_init_value;
  }

  if (!IsProc(t->receiver_index) && !IsProc(t->sender_index)) {
    fprintf(stderr, "DoTransfer: use invalid PTransfer structure\n");
    exit(1);
  }
}

void DoTransferSend(PTransfer t, int32 expected_value, unsigned send_timeout)
{
  Cycle start, end;
  int errc;

  vmmc_async_handle_t msgid;

#if 1
  if (t==NULL || !IsProc(t->sender_index))
    return;
#endif

  assert(t!=NULL);
  assert(IsProc(t->sender_index)); /* this is the sender */

  if (t->sender_async) {
    fprintf(stderr, "DoTransfer: sender: last async has not been finished\n");
    exit(1);
  }
  
  if (expected_value!=((int32*)t->sender_buffer)[t->nwords-1]) {
    fprintf(stderr, "DoTransfer: Sender: expected_value and buffer-end inconsistent\n");
    exit(1);
  }
  if (t->last_init_value==expected_value) {
    fprintf(stderr, "DoTransfer: Sender: init_value is the same as expected_value\n");
    exit(1);
  }
  
  errc=vmmc_SendDataAsync(t->sender_buffer, t->sender_proxy_buffer, t->nwords, &msgid);
  if (errc<0) {
    vmmc_Error(errc, "DoTransfer: Sending error");
    exit(1);
  }
  
  GETCYCLE(start);
  do {
    GETCYCLE(end);
    if (send_timeout!=0 && Interval(start,end)>(float)send_timeout) {
      fprintf(stderr, "DoTransfer: Sender: time out after %d seconds, sending %i to host %d\n",
	      send_timeout, expected_value, t->receiver_index);
      GETCYCLE(start);
    }      
    errc=vmmc_AsyncStatus(msgid);
  } while(errc==vmmcErr_NoSuchError);

#if 0  
  if (mInterval(start,end)>10 && t->nwords<200) {
    fprintf(stderr, "DoTransfer: Sending packet %d(%d words) to host %d takes %.2f ms\n", 
	    expected_value, t->nwords, t->receiver_index, mInterval(start,end));
  }
#endif

  if (errc<0) {
    vmmc_Error(errc, "DoTransfer: Sending error");
    exit(1);
  }
}

void DoTransferReceive(PTransfer t, int32 expected_value, int32 new_init_value, unsigned recv_timeout) 
{
  int32 v;
  Cycle start, end;
#if 1
  if (t==NULL || !IsProc(t->receiver_index))
    return;
#endif

  assert(t!=NULL);
  assert(IsProc(t->receiver_index)); /* this is the sender */

  GETCYCLE(start);
#ifndef INIT_VALUE_CHECK
  /* assume the init value is the first one in the buffer we get, but not expected value */
  v=t->receiver_buffer[t->nwords-1];
  if (v!=expected_value)
    t->last_init_value=v;
#else
  if (t->last_init_value==expected_value) {
    fprintf(stderr, "DoTransfer: Receiver: init_value is the same as expected_value\n");
    exit(1);
  }
#endif
  
  do {
    v=t->receiver_buffer[t->nwords-1];
    GETCYCLE(end);
    if (recv_timeout!=0 && Interval(start,end)>(float)recv_timeout) {
      fprintf(stderr, "DoTransfer: Receiver: time out after %d seconds, waiting %i from host %d\n",
	      recv_timeout, expected_value, t->sender_index);
      GETCYCLE(start);
    }
    
    if (v!=expected_value && v!=t->last_init_value) { /* get a third value */
      fprintf(stderr, "DoTransfer: Receiver: get a value(0x%x) other than expected(0x%x) or init(0x%x) from host %d\n", v, expected_value, t->last_init_value, t->sender_index);
      t->last_init_value=v;
    } 
  } while(v!=expected_value && v==t->last_init_value);
  
  if (v!=expected_value)
    if (v!=t->last_init_value) { /* get a third value */
      fprintf(stderr, "DoTransfer: Receiver: get a value(0x%x) other than expected(0x%x) or init(0x%x)\n", v, expected_value, t->last_init_value);
      exit(1);
    } 
    else { /* receiver time out */
      assert(0);
      fprintf(stderr, "DoTransfer: Receiver: time out after %d seconds\n", recv_timeout);
      exit(1);
    }	
  
  /* get expected value */
  t->receiver_buffer[t->nwords-1]=new_init_value;
  t->last_init_value=new_init_value;
}

void DoTransferGet(PTransfer t, unsigned send_timeout)
{
  Cycle start, end;
  int errc;
  vmmc_async_handle_t msgid;

#if 1
  if (t==NULL || !IsProc(t->sender_index))
    return;
#endif

  assert(t!=NULL);
  assert(IsProc(t->sender_index)); /* this is the sender */

  if (t->sender_async) {
    fprintf(stderr, "DoTransfer: sender: last async has not been finished\n");
    exit(1);
  }
  
  errc=vmmc_GetDataAsync(t->sender_buffer, t->sender_proxy_buffer, t->nwords, &msgid);
  if (errc<0) {
    vmmc_Error(errc, "DoTransfer: Getting error");
    exit(1);
  }
  
  GETCYCLE(start);
  do {
    GETCYCLE(end);
    if (send_timeout!=0 && Interval(start,end)>(float)send_timeout) {
      fprintf(stderr, "DoTransfer: Getter: time out after %d seconds, get from host %d\n",
	      send_timeout, t->receiver_index);
      GETCYCLE(start);
    }      
    errc=vmmc_AsyncStatus(msgid);
  } while(errc==vmmcErr_NoSuchError);
  
  if (errc<0) {
    vmmc_Error(errc, "DoTransfer: Getting error:1");
    exit(1);
  }
}

void DoTransferChange(PTransfer t, int32 init_value, unsigned recv_timeout)
{
  /*  receiver only monitors the change of last word in buffer
   * IMPORTANT: init_value should have the same value 
   *  on both sender and receiver
   */

  if (t==NULL)
    return;
  if (IsProc(t->sender_index)) { /* this is the sender */
    Cycle start, end;
    int errc;
    vmmc_async_handle_t msgid;

    if (t->sender_async) {
      fprintf(stderr, "DoTransferChange: sender: last async has not been finished\n");
      exit(1);
    }
      
    if (init_value==((int32*)t->sender_buffer)[t->nwords-1]) {
      fprintf(stderr, "DoTransferChange: Sender: init_value equals buffer-end\n");
      exit(1);
    }

    errc=vmmc_SendDataAsync(t->sender_buffer, t->sender_proxy_buffer, t->nwords, &msgid);
    if (errc<0) {
      vmmc_Error(errc, "DoTransferChange: Sending error");
      exit(1);
    }

    GETCYCLE(start);
    do {
      GETCYCLE(end);
      if (recv_timeout!=0 && Interval(start,end)>(float)recv_timeout) {
	fprintf(stderr, "DoTransferChange: Sender: time out after %d seconds, sending %i words to host %d\n",
		recv_timeout, t->nwords, t->receiver_index);
	GETCYCLE(start);
      }      
      errc=vmmc_AsyncStatus(msgid);
    } while(errc==vmmcErr_NoSuchError);

#if 0    
    if (mInterval(start,end)>10 && t->nwords<200) {
      fprintf(stderr, "DoTransferChange: Sending packet(%d words) to host %d takes %.2f ms\n", 
	      t->nwords, t->receiver_index, mInterval(start,end));
    }
#endif

    if (errc<0) {
      vmmc_Error(errc, "DoTransferChange: Sending error");
      exit(1);
    }
  }
  else if (IsProc(t->receiver_index)) { /* this is the receiver */
    int32 v;
    Cycle start, end;
    
    GETCYCLE(start);

    do {
      v=t->receiver_buffer[t->nwords-1];
      GETCYCLE(end);
      if (recv_timeout!=0 && Interval(start,end)>(float)recv_timeout) {
	fprintf(stderr, "DoTransferChange: Receiver: time out after %d seconds, waiting change(%i) from host %d\n",
		recv_timeout, init_value, t->sender_index);
	GETCYCLE(start);
      }
    } while(v==init_value);
    
    /* get expected value */
    t->receiver_buffer[t->nwords-1];
    t->last_init_value=t->receiver_buffer[t->nwords-1];
  }
  else {
    fprintf(stderr, "DoTransferChange: use invalid PTransfer structure\n");
    exit(1);
  }
}





#if 0
void DoTransferSenderAsync(PTransfer t)
{
  /* IMPORTANT: expected_value and new_init_value should have the same value 
   *  on both sender and receiver
   */

  if (IsProc(t->sender_index)) { /* this is the sender */
    int errc;

    errc=vmmc_SendDataAsync(t->sender_buffer, t->sender_proxy_buffer, t->nwords);
    if (errc<0) {
      vmmc_Error(errc, "DoTransferSenderAsync: async Sending error");
      exit(1);
    }
    t->SendAsync=1;
  }
}
#endif

void MonitorWord(PTransfer t, unsigned index, int32 expected_value, unsigned timeout)
{
  int32 init_value, v;
  Cycle start, end;

  if (t==NULL)
    return;
  if (!IsProc(t->receiver_index)) 
    return;

  assert(index<t->nwords);

  /* only receiver continue */
  GETCYCLE(start);
  v=t->receiver_buffer[index];
  if (v==expected_value)
    return;
  
  init_value=v;
  do {
    v=t->receiver_buffer[index];
    GETCYCLE(end);
    if (timeout!=0 && Interval(start,end)>(float)timeout) {
      fprintf(stderr, "MoniterWord: Receiver: timeout after %d seconds, waiting %i from host %d\n",
	      timeout, expected_value, t->sender_index);
	  fflush(stderr);
      GETCYCLE(start);
    }
    if (v!=expected_value && v!=init_value) { /* get a third value */
      fprintf(stderr, "MoniterWord: Receiver: get a value(0x%x) other than expected(0x%x) or init(0x%x) from host %d\n", v, expected_value, init_value, t->sender_index);
	  fflush(stderr);
      init_value=v;
    } 
    
  } while(v!=expected_value && v==init_value);

  if (v!=expected_value) { /* get a third value */
    fprintf(stderr, "MoniterWord: Receiver: get a value(0x%x) other than expected(0x%x) or init(0x%x) from host %d\n", v, expected_value, init_value, t->sender_index);
	  fflush(stderr);
    exit(1);
  } 
}
  
  
