/****************************************************************************
				vmmc.h
*****************************************************************************
	header files for VMMC user level communication system
*****************************************************************************/

#ifndef __VMMC_H__
#define __VMMC_H__

#ifdef WIN32 
#include <winsock.h>
#else
/* we really mean linux here */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#endif WIN32


/****************************************************************************
			   macros and flags
*****************************************************************************/
/*
 *              vmmc_ExportRecvBuf flags
 */
#define	BUF_WRITE_ONLY		(0x01<<0)	/* this is the default */
#define BUF_REDIRECTABLE	(0x01<<1)
#define BUF_GLOBAL_SPACE	(0x01<<2)
#define BUF_READWRITE		(0x01<<3)
#define BUF_READONLY		(0x01<<4)


/****************************************************************************
			      data types
*****************************************************************************/
/*
 *      int32:
 *
 *      Type of a 32-bit integer. vmmc_WordSize = 32 bits, which is equal to
 *      the granularity of SHRIMP deliberate updates - 4 bytes currently
 */
#ifdef  WIN32
typedef unsigned int    uint;
typedef unsigned long   ulong;
typedef long            int32;
typedef unsigned long   uint32;
#else
typedef int             int32;
#endif  WIN32


/*
 *      VMMC error return values:
 */
#include "vmmcerror.h"


/*
 *  vmmc_callback_t:
 *
 *      The notification handler. It takes two arguments: last word
 *      virtual address and the last word content.
 *
 */
typedef void (*vmmc_callback_t) (int32 *, int32);

/*
 *      vmmc_exphandle_t:
 */
typedef struct vmmc_exphandle *vmmc_exphandle_t;

typedef struct {
     uint32           index;
     uint32           nwords;
     int32            *proxy;
} vmmc_imphandle_t;

/*
 *      vmmc_perm_t:
 */
typedef uint32 vmmc_perm_t;

/*
 *      vmmc_node_t:
 *
 *      unique node address, currently standard IP address
 */
typedef struct {
     uint32           addr;
} vmmc_node_t;

/*
 *      vmmc_pid_t:
 */
typedef uint32 vmmc_pid_t;


/*
 *      vmmc_async_handle_t:
 */
typedef struct {
     uint32           mid;
     uint32           type;
} vmmc_async_handle_t;


/* note to self: CHECK THIS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
#define vmmc_InvalidPid ((vmmc_pid_t) GLOBAL_BUF_SPACE)


/*****************************************************************************
			 vmmc function calls
 *****************************************************************************/

/*
 *	CALLSTYLE: necessary for building DLLs which use __stdcall by default.
 */
#ifdef  WIN32
#define CALLSTYLE   __cdecl
#else /* linux */
#define CALLSTYLE
#endif  /* WIN32 */


/*****************************************************************************
			VMMC Mapping Functions
 *****************************************************************************/

#ifdef  __cplusplus
extern "C" {
#endif  __cplusplus
        
/* mapping setup functions */
        
vmmc_result_t 
CALLSTYLE
vmmc_ExportRecvBuf(
     int32            *buf,
     uint32           nwords,
     uint32	      rbufid,
     vmmc_callback_t  proc,
     uint32           flags,        /* -redirectableBuf, contiguous bits */
     vmmc_perm_t      perm,         /* used to match export-import reqs */
     vmmc_exphandle_t *bhandle     /* OUT */
);

vmmc_result_t 
CALLSTYLE
vmmc_ImportRecvBuf(
     vmmc_node_t      node,         /* remote node's id    */
     vmmc_pid_t       pid,          /* remote process' id  */
     uint32           bufid,        /* remote buffer's id  */
     vmmc_imphandle_t *handle,      /* filled by this call */
     int32            **proxyAddr,  /* filled by this call */
     uint32           *nwords       /* filled by this call */
);

vmmc_result_t 
CALLSTYLE
vmmc_ImportRecvBufAsync(
     vmmc_node_t      node,         /* remote node's id    */
     vmmc_pid_t       pid,          /* remote process' id  */
     uint32           bufid,        /* remote buffer's id  */
     vmmc_imphandle_t *handle       /* filled by this call */
);

vmmc_result_t 
CALLSTYLE
vmmc_ImportRecvBufStatus(
     vmmc_imphandle_t *handle,      /* updated by this call*/
     int32            **proxyAddr,  /* filled by this call */
     uint32           *nwords       /* filled by this call */
);


vmmc_result_t 
CALLSTYLE
vmmc_UnimportRecvBuf(
     vmmc_imphandle_t handle        /* imported buffer handler */
);

vmmc_result_t 
CALLSTYLE
vmmc_UnexportRecvBuf(
     vmmc_exphandle_t handle        /* exported buffer handler */
);

#ifdef  __cplusplus
}
#endif  __cplusplus




/*****************************************************************************
		       data transfer functions
*****************************************************************************/

#ifdef  __cplusplus
extern "C" {
#endif  __cplusplus

vmmc_result_t 
CALLSTYLE 
vmmc_SendData(
     int32            *localSrcAddr,
     int32            *remoteDestProxyAddr,
     uint32           nwords
);

vmmc_result_t 
CALLSTYLE
vmmc_SendDataNotify(
     int32            *localSrcAddr,
     int32            *remoteDestProxyAddr,
     uint32           nwords
);

vmmc_result_t 
CALLSTYLE
vmmc_SendDataAsync(
     int32               *localSrcAddr,
     int32               *remoteDestProxyAddr,
     uint32              nwords,
     vmmc_async_handle_t *handle
);

vmmc_result_t 
CALLSTYLE
vmmc_SendDataAsyncNotify(
     int32               *localSrcAddr,
     int32               *remoteDestProxyAddr,
     uint32              nwords,
     vmmc_async_handle_t *handle
);

vmmc_result_t 
CALLSTYLE
vmmc_GetData(
     int32            *localDstBuffer, 
     int32            *remoteSrcProxyAddr, 
     uint32           nwords
);

vmmc_result_t 
CALLSTYLE
vmmc_GetDataAsync(
     int32               *localDstBuffer, 
     int32               *remoteSrcProxyAddr, 
     uint32              nwords,
     vmmc_async_handle_t *handle
);

vmmc_result_t 
CALLSTYLE
vmmc_AsyncStatus(
     vmmc_async_handle_t handle
);
#ifdef  __cplusplus
}
#endif  __cplusplus



/*****************************************************************************
			notification functions
*****************************************************************************/
#ifdef  __cplusplus
extern "C" {
#endif  __cplusplus

void 
CALLSTYLE
vmmc_BlockNotifications( void );

vmmc_result_t 
CALLSTYLE
vmmc_UnblockNotifications( void );

#ifdef  __cplusplus
}
#endif  __cplusplus



/*****************************************************************************
		       cluster remote functions
*****************************************************************************/

#ifdef  __cplusplus
extern "C" {
#endif  __cplusplus

        
vmmc_result_t 
CALLSTYLE
vmmc_Spawn(
     char             *execfile,
     char             **argv,
     vmmc_node_t      remoteNode,
     vmmc_pid_t       *childPid
);

vmmc_result_t 
CALLSTYLE
vmmc_Parent(
     vmmc_node_t      *parentNode,
     vmmc_pid_t       *parentPid
);

vmmc_result_t 
CALLSTYLE
vmmc_SessionHosts(
     uint32           *nhosts,
     vmmc_node_t      **hostIds
);

vmmc_result_t 
CALLSTYLE
vmmc_AllHosts(
     uint32           *nhosts,
     vmmc_node_t      **hostIds
);
 
vmmc_pid_t
CALLSTYLE
vmmc_MyPid( void );

vmmc_node_t 
CALLSTYLE
vmmc_MyNode( void );


vmmc_node_t 
CALLSTYLE
vmmc_NameToNode(
     char             *hostname
);


char* 
CALLSTYLE
vmmc_MyHostName(
     void
);


char* 
CALLSTYLE
vmmc_NodeToName(
     vmmc_node_t      node
);

int 
CALLSTYLE
vmmc_EqualNodes(
     vmmc_node_t      node1,
     vmmc_node_t      node2
);
 
  

#ifdef  __cplusplus
}
#endif  __cplusplus




/*****************************************************************************
			redirection functions
*****************************************************************************/

#ifdef  __cplusplus
extern "C" {
#endif  __cplusplus

int
CALLSTYLE
vmmc_DataEnd(
     vmmc_exphandle_t bhandle
);

void 
CALLSTYLE
vmmc_ClearDataEnd(
     vmmc_exphandle_t bhandle
);

vmmc_result_t 
CALLSTYLE
vmmc_PostRedir(
     vmmc_exphandle_t bhandle,
     uint32           redirOffset,
     uint32           numWords,
     int32            *userBuf
);

vmmc_result_t 
CALLSTYLE
vmmc_EndRedir(
     vmmc_exphandle_t bhandle,
     uint32           *numWordsPtr,
     uint32           *firstWordOffsetPtr
);

#ifdef  __cplusplus
}
#endif  __cplusplus




/*****************************************************************************
                       miscellaneous functions
*****************************************************************************/

#ifdef  __cplusplus
extern "C" {
#endif  __cplusplus

        
char *
CALLSTYLE 
vmmc_ErrorStr(
     vmmc_result_t    errcode
);

void
CALLSTYLE 
vmmc_Error(
     vmmc_result_t    errcode,
     const char       *msg
);

void
CALLSTYLE 
vmmc_Version(
     uint32           *major,
     uint32           *minor
);

const int
CALLSTYLE
vmmc_PageSize( void );

const int
CALLSTYLE
vmmc_WordSize( void );
        
void
CALLSTYLE
vmmc_SetDebugLevel( int level );



#ifdef  __cplusplus
}
#endif  __cplusplus


#endif  __VMMC_H__


